"""Molecular database (MDB) class

   * MdbExomol is the MDB for ExoMol

Initial code borrowed from the `Exojax <https://github.com/HajimeKawahara/exojax>`__
code (which you should also have a look at !), by @HajimeKawahara, under MIT License.

"""
import pathlib
import warnings

import numpy as np

# from exojax.spec import hapi, exomolapi, exomol
# from exojax.spec.hitran import gamma_natural as gn
import pandas as pd

try:
    from . import exomolapi
    from .exomol_utils import e2s
except ImportError:  # if local import
    from radis.io import exomolapi
    from radis.io.exomol_utils import e2s

from radis.db.classes import get_molecule_identifier

EXOMOL_URL = "http://www.exomol.com/db/"

from urllib.request import HTTPError, urlopen

from bs4 import BeautifulSoup


def get_exomol_full_isotope_name(molecule, isotope):
    """Get full isotope name for ``molecule`` and ``isotope`` number

    Parameters
    ----------
    molecule: str
    isotope: int
        terrestrial abundance

    Examples
    --------
    ::
        get_exomol_full_isotope_name("CH4", 1)
        >>> '12C-1H4'

    See Also
    --------
    :py:func:`~radis.io.exomol.get_exomol_database_list`"""

    from radis.db.classes import KNOWN_EXOMOL_ISOTOPES_NAMES

    if (molecule, isotope) in KNOWN_EXOMOL_ISOTOPES_NAMES:
        return KNOWN_EXOMOL_ISOTOPES_NAMES[(molecule, isotope)]
    else:
        # Read and convert from HITRAN molecules
        from radis.db.molparam import MolParams

        mp = MolParams()
        return mp.get(molecule, isotope, "isotope_name_exomol")


def get_exomol_database_list(molecule, isotope_full_name):
    """Parse ExoMol website and return list of available databases, and recommended database

    Parameters
    ----------
    molecule: str
    isotope_full_name: str
        isotope full name (ex. ``12C-1H4`` for CH4,1). Get it from
        :py:func:`radis.io.exomol.get_exomol_full_isotope_name`

    Returns
    -------

    Examples
    --------
    ::
        databases, recommended = get_exomol_database_list("CH4", "12C-1H4")
        >>> ['xsec-YT10to10', 'YT10to10', 'YT34to10'], 'YT34to10'

    See Also
    --------
    :py:func:`~radis.io.exomol.get_exomol_full_isotope_name`
    """

    url = f"https://exomol.com/data/molecules/{molecule}/{isotope_full_name}"
    try:
        response = urlopen(url).read()
    except HTTPError as err:
        raise ValueError(f"HTTPError opening url={url}") from err

    soup = BeautifulSoup(
        response, features="lxml"
    )  # make soup that is parse-able by bs

    # Recommended database
    rows = soup.find_all(
        "a", {"class": "list-group-item link-list-group-item recommended"}
    )
    databases_recommended = [r.get_attribute_list("title")[0] for r in rows]

    # All others
    rows = soup.find_all("a", {"class": "list-group-item link-list-group-item"})
    databases = [r.get_attribute_list("title")[0] for r in rows]

    assert len(databases_recommended) <= 1

    databases = databases + databases_recommended

    return databases, databases_recommended[0]


def fetch_exomol(
    molecule,
    database=None,
    local_databases=r"~/.radisdb/exomol/",
    databank_name="EXOMOL-{molecule}",
    isotope=None,
    load_wavenum_min=None,
    load_wavenum_max=None,
    cache=True,
    verbose=True,
    clean_cache_files=True,
    return_local_path=False,
    return_partition_function=False,
):
    """Stream ExoMol file from EXOMOL website. Unzip and build a HDF5 file directly.

    Returns a Pandas DataFrame containing all lines.

    Parameters
    ----------
    molecule: str
        ExoMol molecule
    database: str
        database name. Ex:: ``POKAZATEL`` or ``BT2`` for ``H2O``. See
        :py:data:`~radis.io.exomol.KNOWN_EXOMOL_DATABASE_NAMES`. If ``None`` and
        there is only one database available, use it.
    local_databases: str
        where to create the RADIS HDF5 files. Default ``"~/.radisdb/exomol"``
    databank_name: str
        name of the databank in RADIS :ref:`Configuration file <label_lbl_config_file>`
        Default ``"EXOMOL-{molecule}"``
    isotope: str
        load only certain isotopes : ``'2'``, ``'1,2'``, etc. Default ``1``.
    load_wavenum_min, load_wavenum_max: float (cm-1)
        load only specific wavenumbers.

    Other Parameters
    ----------------
    cache: bool, or ``'regen'``
        if ``True``, use existing HDF5 file. If ``False`` or ``'regen'``, rebuild it.
        Default ``True``.
    verbose: bool
    clean_cache_files: bool
        if ``True`` clean downloaded cache files after HDF5 are created.
    return_local_path: bool
        if ``True``, also returns the path of the local database file.
    return_partition_function: bool
        if ``True``, also returns a :py:class:`~radis.levels.partfunc.PartFuncExoMol` object.

    Returns
    -------
    df: pd.DataFrame
        Line list
        A HDF5 file is also created in ``local_databases`` and referenced
        in the :ref:`RADIS config file <label_lbl_config_file>` with name
        ``databank_name``
    local_path: str
        path of local database file if ``return_local_path``

    Notes
    -----
    if using ``load_only_wavenum_above/below`` or ``isotope``, the whole
    database is anyway downloaded and uncompressed to ``local_databases``
    fast access .HDF5 files (which will take a long time on first call). Only
    the expected wavenumber range & isotopes are returned. The .HFD5 parsing uses
    :py:func:`~radis.io.hdf5.hdf2df`

    See Also
    --------
    :py:func:`~radis.io.hdf5.hdf2df`

    """
    full_molecule_name = get_exomol_full_isotope_name(molecule, isotope)
    known_exomol_databases, recommended_database = get_exomol_database_list(
        molecule, full_molecule_name
    )

    _exomol_use_hint = f"Select one of them with `radis.fetch_exomol(DATABASE_NAME)`, `SpectrumFactory.fetch_databank('exomol', exomol_database=DATABASE_NAME')`, or `calc_spectrum(..., databank=('exomol', DATABASE_NAME))` \n"
    if database is None:
        if len(known_exomol_databases) == 1:
            database = known_exomol_databases[0]
            if verbose:
                print(f"Using ExoMol database {database} for {full_molecule_name}")
        elif recommended_database:
            database = recommended_database
            if verbose:
                print(
                    f"Using ExoMol database {database} for {full_molecule_name} (recommended by the ExoMol team). All available databases are {known_exomol_databases}. {_exomol_use_hint}"
                )
        else:
            raise KeyError(
                f"Choose one of the several databases available for {full_molecule_name} in ExoMol: {known_exomol_databases}. ({recommended_database} is recommended by the ExoMol team). {_exomol_use_hint}"
            )
    else:
        if database not in known_exomol_databases:
            raise KeyError(
                f"{database} is not of the known available ExoMol databases for {full_molecule_name}. Choose one of : {known_exomol_databases}. ({recommended_database} is recommended by the ExoMol team). {_exomol_use_hint}"
            )

    local_path = (
        pathlib.Path(local_databases).expanduser()
        / molecule
        / full_molecule_name
        / database
    )

    mdb = MdbExomol(local_path, [load_wavenum_min, load_wavenum_max])

    df = mdb.to_df()

    out = [df]

    if return_local_path:
        out.append(str(mdb.path))
    if return_partition_function:
        assert return_local_path
        out.append(mdb.to_partition_function_tabulator())

    return out


class MdbExomol(object):
    """molecular database of ExoMol

    MdbExomol is a class for ExoMol.

    Attributes:
        nu_lines (nd array): line center (cm-1)
        Sij0 (nd array): line strength at T=Tref (cm)
        dev_nu_lines (np array): line center in device (cm-1)
        logsij0 (np array): log line strength at T=Tref
        A (np array): Einstein A coefficient
        elower (np array): the lower state energy (cm-1)
        gpp (np array): statistical weight
        jlower (np array): J_lower
        jupper (np array): J_upper
        n_Tref (np array): temperature exponent
        alpha_ref (np array): alpha_ref (gamma0)
        n_Tref_def: default temperature exponent in .def file, used for jlower not given in .broad
        alpha_ref_def: default alpha_ref (gamma0) in .def file, used for jlower not given in .broad

    """

    def __init__(
        self,
        path,
        nurange=[-np.inf, np.inf],
        margin=1.0,
        crit=-np.inf,
        bkgdatm="H2",
        broadf=True,
    ):
        """Molecular database for Exomol form

        Args:
           path: path for Exomol data directory/tag. For instance, "/home/CO/12C-16O/Li2015"
           nurange: wavenumber range list (cm-1) or wavenumber array
           margin: margin for nurange (cm-1)
           crit: line strength lower limit for extraction
           bkgdatm: background atmosphere for broadening. e.g. H2, He,
           broadf: if False, the default broadening parameters in .def file is used

        Note:
           The trans/states files can be very large. For the first time to read it, we convert it to the feather-format. After the second-time, we use the feather format instead.

        """
        explanation = "Note: Couldn't find the feather format. We convert data to the feather format. After the second time, it will become much faster."

        self.path = pathlib.Path(path)
        t0 = self.path.parents[0].stem
        molec = t0 + "__" + str(self.path.stem)
        self.bkgdatm = bkgdatm
        print("Background atmosphere: ", self.bkgdatm)
        molecbroad = t0 + "__" + self.bkgdatm

        self.crit = crit
        self.margin = margin
        self.nurange = [np.min(nurange), np.max(nurange)]
        self.broadf = broadf
        # Where exomol files are
        self.states_file = self.path / pathlib.Path(molec + ".states.bz2")
        self.pf_file = self.path / pathlib.Path(molec + ".pf")
        self.def_file = self.path / pathlib.Path(molec + ".def")
        self.broad_file = self.path / pathlib.Path(molecbroad + ".broad")

        if not self.def_file.exists():
            self.download(molec, extension=[".def"])
        if not self.pf_file.exists():
            self.download(molec, extension=[".pf"])
        if not self.states_file.exists():
            self.download(molec, extension=[".states.bz2"])
        if not self.broad_file.exists():
            self.download(molec, extension=[".broad"])

        # Add molecule name
        tag = molec.split("__")
        self.isotope_fullname = tag[0]
        self.molecule = e2s(tag[0])
        self.isotope = 1  # Placeholder. TODO : impement parsing of other isotopes.

        # load def
        # @minou: I feel this next line should just update the self.stuff
        dic_def = exomolapi.read_def(self.def_file)  # approx. 3 ms
        self.n_Texp_def = dic_def["n_Texp"]
        self.alpha_ref_def = dic_def["alpha_ref"]
        self.molmass = dic_def["molmass"]

        #  default n_Texp value if not given
        if self.n_Texp_def is None:
            warnings.warn(
                Warning(
                    """
                    No default broadening exponent in def file. Assigned n = 0.5
                    """
                )
            )
            self.n_Texp_def = 0.5
        #  default alpha_ref value if not given
        if self.alpha_ref_def is None:
            warnings.warn(
                Warning(
                    """
                    No default broadening in def file. Assigned alpha_ref = 0.07
                    """
                )
            )
            self.alpha_ref_def = 0.07

        # load states
        if self.states_file.with_suffix(".feather").exists():
            states = pd.read_feather(self.states_file.with_suffix(".feather"))
        else:
            print(explanation)
            states = exomolapi.read_states(self.states_file, dic_def)
            states.to_feather(self.states_file.with_suffix(".feather"))
        # load pf
        pf = exomolapi.read_pf(self.pf_file)
        self.gQT = pf["QT"].to_numpy()  # grid QT
        self.T_gQT = pf["T"].to_numpy()  # T forgrid QT

        # trans file(s)
        print("Reading transition file")
        if dic_def["numinf"] is None:
            self.trans_file = self.path / pathlib.Path(molec + ".trans.bz2")
            if not self.trans_file.exists():
                self.download(molec, [".trans.bz2"])

            if self.trans_file.with_suffix(".feather").exists():
                trans = pd.read_feather(self.trans_file.with_suffix(".feather"))
            else:
                print(explanation)
                trans = exomolapi.read_trans(self.trans_file)
                trans.to_feather(self.trans_file.with_suffix(".feather"))

            #!!TODO:restrict NOW the trans size to avoid useless overload of memory and CPU
            # trans = trans[(trans['nu'] > self.nurange[0] - self.margin) & (trans['nu'] < self.nurange[1] + self.margin)]
            # compute gup and elower
            (
                self._A,
                self.nu_lines,
                self._elower,
                self._gpp,
                self._jlower,
                self._jupper,
                self._quantumNumbers,
            ) = exomolapi.pickup_gE(states, trans, dic_def)
        else:
            imin = (
                np.searchsorted(dic_def["numinf"], nurange[0], side="right") - 1
            )  # left side
            imax = (
                np.searchsorted(dic_def["numinf"], nurange[1], side="right") - 1
            )  # left side
            self.trans_file = []
            for k, i in enumerate(range(imin, imax + 1)):
                trans_file = self.path / pathlib.Path(
                    molec + "__" + dic_def["numtag"][i] + ".trans.bz2"
                )
                if not trans_file.exists():
                    self.download(
                        molec, extension=[".trans.bz2"], numtag=dic_def["numtag"][i]
                    )
                if trans_file.with_suffix(".feather").exists():
                    trans = pd.read_feather(trans_file.with_suffix(".feather"))
                else:
                    print(explanation)
                    trans = exomolapi.read_trans(trans_file)
                    trans.to_feather(trans_file.with_suffix(".feather"))
                    #!!TODO:restrict NOW the trans size to avoid useless overload of memory and CPU
                    # trans = trans[(trans['nu'] > self.nurange[0] - self.margin) & (trans['nu'] < self.nurange[1] + self.margin)]

                self.trans_file.append(trans_file)

                # compute gup and elower
                if k == 0:
                    (
                        self._A,
                        self.nu_lines,
                        self._elower,
                        self._gpp,
                        self._jlower,
                        self._jupper,
                        self._quantumNumbers,
                    ) = exomolapi.pickup_gE(states, trans, dic_def)
                else:
                    (
                        Ax,
                        nulx,
                        elowerx,
                        gppx,
                        jlowerx,
                        jupperx,
                        _quantumNumbersx,
                    ) = exomolapi.pickup_gE(states, trans, dic_def)
                    self._A = np.hstack([self._A, Ax])
                    self.nu_lines = np.hstack([self.nu_lines, nulx])
                    self._elower = np.hstack([self._elower, elowerx])
                    self._gpp = np.hstack([self._gpp, gppx])
                    self._jlower = np.hstack([self._jlower, jlowerx])
                    self._jupper = np.hstack([self._jupper, jupperx])
                    self._quantumNumbers = np.hstack(
                        [self._quantumNumbers, _quantumNumbersx]
                    )

        self.Tref = 296.0
        self.QTref = np.array(self.QT_interp(self.Tref))

        Ia = 1  #  TODO    Add isotope abundance

        from radis.lbl.base import linestrength_from_Einstein  # TODO: move elsewhere

        self.Sij0 = linestrength_from_Einstein(
            A=self._A,
            gu=self._gpp,
            El=self._elower,
            Ia=Ia,
            nu=self.nu_lines,
            Q=self.QTref,
            T=self.Tref,
        )

        ### MASKING ###
        mask = (
            (self.nu_lines > self.nurange[0] - self.margin)
            * (self.nu_lines < self.nurange[1] + self.margin)
            * (self.Sij0 > self.crit)
        )

        self.masking(mask)

    def masking(self, mask):
        """applying mask and (re)generate jnp.arrays

        Args:
           mask: mask to be applied. self.mask is updated.

        """
        # TODO : replace with HDF5 masking ?

        # numpy float 64 Do not convert them jnp array
        self.nu_lines = self.nu_lines[mask]
        self.Sij0 = self.Sij0[mask]
        self._A = self._A[mask]
        self._elower = self._elower[mask]
        self._gpp = self._gpp[mask]
        self._jlower = self._jlower[mask]
        self._jupper = self._jupper[mask]

        # jnp arraysgamma_
        self.dev_nu_lines = np.array(self.nu_lines)
        self.logsij0 = np.array(np.log(self.Sij0))
        self.A = np.array(self._A)
        # self.gamma_natural = gn(self.A)   #  Natural Broadening neglected in RADIS
        self.elower = np.array(self._elower)
        self.gpp = np.array(self._gpp)
        self.jlower = np.array(self._jlower, dtype=int)
        self.jupper = np.array(self._jupper, dtype=int)
        ##Broadening parameters
        self.set_broadening()

        for key, array in self._quantumNumbers.items():
            self._quantumNumbers[key] = array[mask]

    def set_broadening(self, alpha_ref_def=None, n_Texp_def=None):
        """setting broadening parameters

        Args:
           alpha_ref: set default alpha_ref and apply it. None=use self.alpha_ref_def
           n_Texp_def: set default n_Texp and apply it. None=use self.n_Texp_def
        """
        if alpha_ref_def:
            self.alpha_ref_def = alpha_ref_def
        if n_Texp_def:
            self.n_Texp_def = n_Texp_def

        if self.broadf:
            try:
                print(".broad is used.")
                bdat = exomolapi.read_broad(self.broad_file)
                codelv = exomolapi.check_bdat(bdat)
                print("Broadening code level=", codelv)
                if codelv == "a0":
                    j2alpha_ref, j2n_Texp = exomolapi.make_j2b(
                        bdat,
                        alpha_ref_default=self.alpha_ref_def,
                        n_Texp_default=self.n_Texp_def,
                        jlower_max=np.max(self._jlower),
                    )
                    self.alpha_ref = np.array(j2alpha_ref[self._jlower])
                    self.n_Texp = np.array(j2n_Texp[self._jlower])
                elif codelv == "a1":
                    j2alpha_ref, j2n_Texp = exomolapi.make_j2b(
                        bdat,
                        alpha_ref_default=self.alpha_ref_def,
                        n_Texp_default=self.n_Texp_def,
                        jlower_max=np.max(self._jlower),
                    )
                    jj2alpha_ref, jj2n_Texp = exomolapi.make_jj2b(
                        bdat,
                        j2alpha_ref_def=j2alpha_ref,
                        j2n_Texp_def=j2n_Texp,
                        jupper_max=np.max(self._jupper),
                    )
                    self.alpha_ref = np.array(jj2alpha_ref[self._jlower, self._jupper])
                    self.n_Texp = np.array(jj2n_Texp[self._jlower, self._jupper])
            except:
                print(
                    "Warning: Cannot load .broad. The default broadening parameters are used."
                )
                self.alpha_ref = np.array(
                    self.alpha_ref_def * np.ones_like(self._jlower)
                )
                self.n_Texp = np.array(self.n_Texp_def * np.ones_like(self._jlower))

        else:
            print("The default broadening parameters are used.")
            self.alpha_ref = np.array(self.alpha_ref_def * np.ones_like(self._jlower))
            self.n_Texp = np.array(self.n_Texp_def * np.ones_like(self._jlower))

    def QT_interp(self, T):
        """interpolated partition function

        Args:
           T: temperature

        Returns:
           Q(T) interpolated in jnp.array

        """
        return np.interp(T, self.T_gQT, self.gQT)

    def qr_interp(self, T):
        """interpolated partition function ratio

        Args:
           T: temperature

        Returns:
           qr(T)=Q(T)/Q(Tref) interpolated in jnp.array

        """
        return self.QT_interp(T) / self.QT_interp(self.Tref)

    def download(self, molec, extension, numtag=None):
        """Downloading Exomol files

        Args:
           molec: like "12C-16O__Li2015"
           extension: extension list e.g. [".pf",".def",".trans.bz2",".states.bz2",".broad"]
           numtag: number tag of transition file if exists. e.g. "11100-11200"

        Note:
           The download URL is written in exojax.utils.url.

        """
        import os
        import urllib.request

        tag = molec.split("__")
        molname_simple = e2s(tag[0])

        for ext in extension:
            if ext == ".trans.bz2" and numtag is not None:
                ext = "__" + numtag + ext

            if ext == ".broad":
                pfname_arr = [
                    tag[0] + "__H2" + ext,
                    tag[0] + "__He" + ext,
                    tag[0] + "__air" + ext,
                ]
                url = EXOMOL_URL + molname_simple + "/" + tag[0] + "/"
            else:
                pfname_arr = [molec + ext]
                url = EXOMOL_URL + molname_simple + "/" + tag[0] + "/" + tag[1] + "/"

            for pfname in pfname_arr:
                pfpath = url + pfname
                os.makedirs(str(self.path), exist_ok=True)
                print("Downloading " + pfpath)
                try:
                    urllib.request.urlretrieve(pfpath, str(self.path / pfname))
                except:
                    print("Error: Couldn't download " + ext + " file and save.")

    def to_partition_function_tabulator(self):
        from radis.levels.partfunc import PartFuncExoMol

        return PartFuncExoMol(self.isotope_fullname, self.T_gQT, self.gQT)

    def to_df(self):
        """Export Line Database to a RADIS-friendly Pandas DataFrame"""

        ##Broadening parameters
        self.set_broadening()

        # TODO : define RADIS database format somewhere else; with description of the column names.
        df = pd.DataFrame(
            {
                "wav": self.nu_lines,
                "int": self.Sij0,
                "A": self._A,
                "airbrd": self.alpha_ref,  # temperature dependance exponent for air broadening
                # "selbrd": None,                  # self-temperature dependant exponent. Not implementedi in ExoMol
                "El": self._elower,
                # "Tdpair": None,
                # "Pshft": None,
                "ju": self._jupper,
                "jl": self._jlower,
                "gpp": self._gpp,
                "Tdpair": self.n_Texp,  # temperature dependance exponent. Here we use Tdair; no Tdpsel. TODO
            }
        )
        for key, array in self._quantumNumbers.items():
            try:
                df[key] = pd.to_numeric(array, errors="raise")
            except:
                df[key] = array
                df[key] = df[key].astype("str")

        from radis.db.classes import HITRAN_MOLECULES

        if self.molecule in HITRAN_MOLECULES:
            df.attrs["id"] = get_molecule_identifier(self.molecule)
        df.attrs["molecule"] = self.molecule
        df.attrs["iso"] = 1  # TODO : get isotope number while parsing ExoMol name

        assert self.Tref == 296  # default in RADIS

        return df


if __name__ == "__main__":
    # mdb=MdbExomol("/home/kawahara/exojax/data/CO/12C-16O/Li2015/")
    # mdb=MdbExomol("/home/kawahara/exojax/data/CH4/12C-1H4/YT34to10/",nurange=[6050.0,6150.0])

    # %% Test  by overriding Spectrumfactory's DataFrame df0
    # mdb = MdbExomol(".database/H2O/1H2-16O/POKAZATEL", [4310.0, 4320.0], crit=1.0e-45)
    # df = mdb.to_df()
    # from radis import SpectrumFactory
    # sf = SpectrumFactory(
    #     4310,
    #     4320,
    #     molecule="H2O",
    #     isotope="1",
    # )
    # sf.fetch_databank(
    #     "hitran"
    # )  # placeholder. Load lines (will be replaced), load partition function.
    # sf.df0 = df  # override.
    # s = sf.eq_spectrum(500, name="ExoMol")
    # # sf.fetch_databank('hitran')  # placeholder. Load lines (will be replaced), load partition function.
    # # s_hit = sf.eq_spectrum(500, name='HITRAN')

    #%% Test by direct caclulation
    from radis import SpectrumFactory

    sf = SpectrumFactory(
        4310,
        4320,
        molecule="H2O",
        isotope="1",
        verbose=3,
    )
    sf.fetch_databank("exomol")
    s = sf.eq_spectrum(500, name="ExoMol")
    s.plot()

#    mask=mdb.A>1.e-42
#    mdb.masking(mask)
#    mdb=MdbExomol("/home/kawahara/exojax/data/exomol/NH3/14N-1H3/CoYuTe/",nurange=[6050.0,6150.0])
#    mdb=MdbExomol("/home/kawahara/exojax/data/exomol/H2S/1H2-32S/AYT2/",nurange=[6050.0,6150.0])
#    mdb=MdbExomol("/home/kawahara/exojax/data/exomol/FeH/56Fe-1H/MoLLIST/",nurange=[6050.0,6150.0])
#    mdb=MdbExomol("/home/kawahara/exojax/data/exomol/NO/14N-16O/NOname/14N-16O__NOname")



# -*- coding: utf-8 -*-
"""

::

                *(((((((
                 ((((((((((((              ,(((((
                 ((((((((((((((((/   *((((((((((*
                  ((((((((((((((((( ((((((((((((
                      (((((((( (((((((((((((
                         *
                       @@  *@@       ..  /@@
                  @@&  @@  *@@       @@  /@@  @@%
              @@  @@&  @@  *@@  @@&  @@  /@@  @@%  @@
              @@  @@&  @@  *@@  @@&  @@  /@@  @@%  @@
              @@  @@&  @@  *@@  @@&  @@  /@@  @@%  @@  (@
         ,@   @@  @@&  @@  *@@  @@&  @@  /@@  @@%  @@
         @@   @@  @@&  @@  ,.
                                    ,%&&&&&&&&&&&&&&&&&&&
          &&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&
           &&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&
             &&&&&&&&&&&&&&&&@@@@@@&@@@&&&@@@&&&&&&&&
               &&&&&&&&&&&&&&&@@@@@@&&&&&&&&&&&&&&&
                 &&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&
                   &&&&&&&&&&&&&&&&&&&&&&&&&&&.
                       &&&&&&&&&&&&&&&&&&&
                               .**.
                                &&&,
                                 &&

"""

import os

from .misc.config import get_config
from .misc.utils import Chdir as _chdir
from .misc.utils import getProjectRoot

# %% Config files

# @dev: refactor in progress.
# So far there are config files in ~/radis.json (for databanks), global variables
# here, and a radis/config.json file.
# Everything should be merged in a user JSON file ~/radis.json (json) overriding
# the default one.

config = get_config()
"""dict: RADIS configuration parameters

Parameters
----------
"DEBUG_MODE":False

    bool: change this at runtime with::

        import radis
        radis.config["DEBUG_MODE"] = True

    Use the :py:func:`~radis.misc.debug.printdbg` function in ``radis.misc``, typically with::

        if __debug__: printdbg(...)

    so that printdbg are removed by the Python preprocessor when running in
    optimize mode::

        python -O *.py

"AUTO_UPDATE_SPEC": False

    bool: experimental feature
    used to autoupdate .spec files to the latest format, by simply saving
    them again once they're loaded and fixed.
    Warning! Better have a copy of your files before that, or a way to regenerate
    them.

    Example : Add to the top of your script (once is enough!)::

        import radis
        radis.config["AUTO_UPDATE_SPEC"] = True

    See Also
    --------
    :py:func:`~radis.tools.database._update_to_latest_format`


"AUTO_UPDATE_DATABASE": False
	bool: experimental feature
	used to autoupdate .h5 files to the latest format, by simply saving
	them again once they're loaded and fixed.
	Warning! Better have a copy of your files before that, or a way to regenerate
	them.

	Examples
	--------

	Add to the top of your script (once is enough!)::

	    import radis
	    radis.AUTO_UPDATE_DATABASE = True

	See Also
	--------
	:py:func:`~radis.io.hdf5.hdf2df`


"OLDEST_COMPATIBLE_VERSION": "0.9.1"
    str: forces to regenerate cache files that were created in a previous version

    See Also
    --------
    :py:func:`~radis.io.cache_files.load_h5_cache_file`


"USE_CYTHON": True
    bool: try to use Cython functions when possible

    See Also
    --------
    :py:func:`~radis.misc.arrays.add_at`


"GRIDPOINTS_PER_LINEWIDTH_WARN_THRESHOLD": 3
    float: to determine the optimal value
    of wstep using minimum FWHM value of spectrum.
    Makes sure there are enough gridpoints per line.

    See Also
    --------
    :py:meth:`~radis.lbl.broadening.BroadenFactory._check_accuracy`


"GRIDPOINTS_PER_LINEWIDTH_ERROR_THRESHOLD": 1
    float: to determine the minimum feasible value
    of wstep using minimum FWHM value of spectrum.
    Makes sure there are enough gridpoints per line.

    See Also
    --------
    :py:meth:`~radis.lbl.broadening.BroadenFactory._check_accuracy`


Notes
-----

Default values are read from the ``radis/config.json`` file.

All values are overriden at runtime by the keys in the user JSON file ``~/radis.json (json)``
(in particular, the list of databases)
"""
# TODO : Refactor in progress.


# %% Version


def get_version(verbose=False, add_git_number=True):
    """Reads `__version.txt__
    <https://github.com/radis/radis/blob/master/radis/__version__.txt>`__ and
    retrieve version number. If ``add_git_number``, also appends Git commit
    number if we're on a gitted session.

    Examples
    --------

    ::

        import radis
        print(radis.get_version())
        >>> '0.9.17'
    """

    # First get version
    with open(os.path.join(getProjectRoot(), "__version__.txt")) as version_file:
        version = version_file.read().strip()

    # Now get git info
    if add_git_number:
        import subprocess
        import sys

        cd = _chdir(os.path.dirname(__file__))
        try:
            label = subprocess.check_output("git describe", stderr=subprocess.DEVNULL)
        except:
            if verbose:
                print("couldnt get git version: {0}".format(sys.exc_info()[1]))
            # probably not a git session. drop
        else:
            commit = label.decode().strip().split("-")[-1]
            version = version + "-" + commit
        finally:
            cd.__del__()

    return version


__version__ = get_version(add_git_number=False)
version = get_version(add_git_number=False)


# %% Global namespace

__all__ = [
    "config",
    "version",
    "__version__",
]

# prevent cyclic importants:
from . import db, io, lbl, los, misc, phys, spectrum, tools
from .db import *  # database of molecules
from .io import *  # input / output
from .lbl import *  # line-by-line module
from .levels import *  # rovibrational energies and partition functions
from .los import *  # line-of-sight module
from .phys import *  # conversion functions, blackbody objects
from .spectrum import *  # Spectrum object
from .test import *  # test
from .tools import *  # slit, database, line survey, etc.

__all__.extend(db.__all__)
__all__.extend(io.__all__)
__all__.extend(lbl.__all__)
__all__.extend(los.__all__)
__all__.extend(phys.__all__)
__all__.extend(spectrum.__all__)
__all__.extend(tools.__all__)



# -*- coding: utf-8 -*-
"""
Created on Tue Jul 27 02:04:14 2021

@author: erwan
"""

import os
import os.path
import time


def test_perf_profile(*args, **kwargs):
    """Test visual/interactive performance profile

    Examples
    --------
    ::
        s = calc_spectrum(...)
        s.print_perf_profile()

        # output >>
            spectrum_calculation      0.189s ████████████████
                check_line_databank              0.000s
                check_non_eq_param               0.042s ███
                fetch_energy_5                   0.015s █
                calc_weight_trans                0.008s
                reinitialize                     0.002s
                    copy_database                    0.000s
                    memory_usage_warning             0.002s
                    reset_population                 0.000s
                calc_noneq_population            0.041s ███
                    part_function                    0.035s ██
                    population                       0.006s
                scaled_non_eq_linestrength       0.005s
                    map_part_func                    0.001s
                    corrected_population_se          0.003s
                calc_emission_integral           0.006s
                applied_linestrength_cutoff      0.002s
                calc_lineshift                   0.001s
                calc_hwhm                        0.007s
                generate_wavenumber_arrays       0.001s
                calc_line_broadening             0.074s ██████
                    precompute_DLM_lineshapes        0.012s
                    DLM_Initialized_vectors          0.000s
                    DLM_closest_matching_line        0.001s
                    DLM_Distribute_lines             0.001s
                    DLM_convolve                     0.060s █████
                    others                           0.001s
                calc_other_spectral_quan         0.003s
                generate_spectrum_obj            0.000s
                others                           -0.016s
    ::
        s.generate_perf_profile()

    See output of :py:meth:`~radis.spectrum.spectrum.Spectrum.generate_perf_profile`
    in https://github.com/radis/radis/pull/325

    See Also
    --------
    :py:meth:`~radis.spectrum.spectrum.Spectrum.print_perf_profile`,
    :py:meth:`~radis.spectrum.spectrum.Spectrum.generate_perf_profile`
    """
    from radis.lbl.calc import calc_spectrum
    from radis.test.utils import setup_test_line_databases

    setup_test_line_databases()

    s = calc_spectrum(
        2000,
        2300,
        molecule="CO",
        Tvib=3000,
        Trot=2000,
        databank="HITRAN-CO-TEST",
        verbose=3,
    )

    # Test Console profiler
    s.print_perf_profile()

    # Test interactive profiler
    if os.path.exists("spectrum.prof"):
        os.remove("spectrum.prof")
    s.generate_perf_profile()
    time.sleep(
        5
    )  # note @dev: would be better to wait for the subprocess to end; however it won't even if the window is closed
    with open("spectrum.prof", "rb") as f:
        assert "calculation_time" in str(f.read())

    os.remove("spectrum.prof")


def test_perf_profile_from_factory(*args, **kwargs):
    """See :py:func:`radis.test.spectrum.test_utils.test_perf_profile`"""

    from radis import SpectrumFactory
    from radis.test.utils import setup_test_line_databases

    setup_test_line_databases()

    sf = SpectrumFactory(
        2000,
        2300,
        molecule="CO",
        verbose=3,
    )
    sf.load_databank("HITRAN-CO-TEST")
    sf.non_eq_spectrum(3000, 2000)

    # Test Console profiler
    sf.print_perf_profile()

    # Test interactive profiler
    if os.path.exists("spectrum.prof"):
        os.remove("spectrum.prof")
    sf.generate_perf_profile()
    time.sleep(
        5
    )  # note @dev: would be better to wait for the subprocess to end; however it won't even if the window is closed
    with open("spectrum.prof", "rb") as f:
        assert "calculation_time" in str(f.read())

    os.remove("spectrum.prof")


if __name__ == "__main__":
    test_perf_profile()
    test_perf_profile_from_factory()



# -*- coding: utf-8 -*-
"""
Created on Sun Aug  5 14:26:44 2018

@author: erwan
"""

import numpy as np
import pytest

from radis.los import MergeSlabs, SerialSlabs
from radis.spectrum.compare import get_diff, plot_diff
from radis.spectrum.operations import (
    Radiance,
    Radiance_noslit,
    Transmittance_noslit,
    add_constant,
    crop,
    multiply,
    offset,
    sub_baseline,
)
from radis.test.utils import getTestFile
from radis.tools.database import load_spec


@pytest.mark.fast
def test_crop(verbose=True, *args, **kwargs):
    """Test that update can correctly recompute missing quantities"""

    # 1) A crop example in the same unit as the one stored

    # Work with a Spectrum object that was generated by Specair
    s = load_spec(getTestFile("N2C_specair_380nm.spec"))
    # Focus on N2(C->B), v'=0, v''=2 band:
    s.crop(376, 381, "nm")

    w = s.get_wavelength()
    assert w.min() >= 360
    assert w.max() <= 382

    # 2) A crop example in different unit as the one stored

    # Work with a Spectrum object that was generated by Specair
    s1 = load_spec(getTestFile("CO_Tgas1500K_mole_fraction0.01.spec"), binary=True)
    # Focus on N2(C->B), v'=0, v''=2 band:
    s1.crop(4530, 4533, "nm")

    w = s1.get_wavelength()
    assert w.min() >= 4530
    assert w.max() <= 4533

    return True


@pytest.mark.fast
def test_cut_recombine(verbose=True, *args, **kwargs):
    """
    Use :func:`~radis.spectrum.operations.crop` and :func:`~radis.los.slabs.MergeSlabs`
    to cut a Spectrum and recombine it

    Assert we still get the same spectrum at the end
    """

    s = load_spec(getTestFile("CO_Tgas1500K_mole_fraction0.01.spec"), binary=True)
    # Cut in half
    cut = 2177
    s1 = crop(s, 2000, cut, "cm-1", inplace=False)
    s2 = crop(s, cut, 2300, "cm-1", inplace=False)

    # Recombine
    s_new = MergeSlabs(s1, s2, resample="full", out="transparent")

    # Compare
    assert s.compare_with(s_new, spectra_only=True, plot=False, verbose=verbose)


@pytest.mark.fast
def test_invariants(*args, **kwargs):
    """Ensures adding 0 or multiplying by 1 does not change the spectra"""
    from radis import load_spec
    from radis.test.utils import getTestFile

    s = load_spec(getTestFile("CO_Tgas1500K_mole_fraction0.01.spec"))
    s.update()
    s = Radiance_noslit(s)

    assert s.compare_with(
        add_constant(s, 0, "W/cm2/sr/nm"), plot=False, spectra_only="radiance_noslit"
    )
    assert s.compare_with(multiply(s, 1), plot=False, spectra_only="radiance_noslit")

    assert 3 * s / 3 == s
    assert (1 + s) - 1 == s


@pytest.mark.fast
def test_operations_inplace(verbose=True, *args, **kwargs):

    from radis.spectrum.operations import Radiance_noslit

    s = load_spec(getTestFile("CO_Tgas1500K_mole_fraction0.01.spec"), binary=True)
    s.update("radiance_noslit", verbose=False)
    s = Radiance_noslit(s)

    # Add 1, make sure it worked
    I_max = s.get("radiance_noslit")[1].max()
    s += 1
    assert s.get("radiance_noslit")[1].max() == I_max + 1
    if verbose:
        print("test_operations: s += 1: OK")

    # Multiply, make sure it worked
    I_max = s.get("radiance_noslit")[1].max()
    s *= 10
    assert s.get("radiance_noslit")[1].max() == 10 * I_max
    if verbose:
        print("test_operations: s *= 10: OK")


@pytest.mark.fast
def test_serial_operator(verbose=True, plot=False, *args, **kwargs):
    import matplotlib.pyplot as plt

    s = load_spec(getTestFile("CO_Tgas1500K_mole_fraction0.01.spec"))
    s.update()
    s1 = s.rescale_path_length(10, inplace=False)  # make non optically thin
    s2 = s.rescale_path_length(20, inplace=False)  # make non optically thin
    s3 = s.rescale_path_length(30, inplace=False)  # make non optically thin
    if plot:
        #        s.plot('radiance_noslit', lw=2, nfig='Line of sight (SerialSlabs): s > s > s')
        plt.figure("test_serial_operator").clear()
        SerialSlabs(s1, s2, s3).apply_slit(1).plot(nfig="same")
        ((s1 > s2) > s3).apply_slit(1).plot(nfig="same")
    assert SerialSlabs(s1, s2, s3).compare_with(
        (s1 > s2) > s3, spectra_only=True, plot=False
    )
    assert (s1 > (s2 > s3)) == ((s1 > s2) > s3)
    # Forbidden syntax:
    with pytest.raises(ArithmeticError):
        assert SerialSlabs(s1, s2, s3).compare_with(
            s1 > s2 > s3, spectra_only=True, plot=False
        )


@pytest.mark.fast
def test_multiplyAndAddition(verbose=True, plot=False, *args, **kwargs):

    s = load_spec(getTestFile("CO_Tgas1500K_mole_fraction0.01.spec"), binary=True)
    s.update("radiance_noslit", verbose=False)
    s.apply_slit(0.1)
    s = Radiance(s)
    assert s.units["radiance"] == "mW/cm2/sr/nm"

    s_bis = add_constant(s, 1, "mW/cm2/sr/nm")
    w, Idiff = get_diff(s_bis, s, "radiance")
    test = Idiff[1] - 1
    assert np.all(test < 1e-10)

    s_ter = multiply(multiply(s, 50), 1 / 50)
    #    plot_diff(s_ter, s_5)
    diff = get_diff(s_ter, s, "radiance")
    ratio = abs(np.trapz(diff[1], x=diff[0]) / s.get_integral("radiance"))
    assert ratio < 1e-10


@pytest.mark.fast
def test_offset(verbose=True, plot=False, *args, **kwargs):

    s = load_spec(getTestFile("CO_Tgas1500K_mole_fraction0.01.spec"), binary=True)
    s.update("radiance_noslit", verbose=False)
    s.apply_slit(0.1)

    s2 = offset(s, 10, "nm", name="offset_10nm")
    if plot:
        plot_diff(s, s2)
    assert np.allclose(s2.get_wavelength(), s.get_wavelength() + 10)

    # Test inplace version
    s.offset(10, "nm")
    assert np.allclose(s2.get_wavelength(), s.get_wavelength())


@pytest.mark.fast
def test_other_algebraic_operations(verbose=True, plot=False, *args, **kwargs):

    # An implement of Spectrum Algebra
    # Reload:
    s = load_spec(getTestFile("CO_Tgas1500K_mole_fraction0.01.spec"))
    s.update()

    # Test addition of Spectra
    s.plot(lw=2, nfig="Merge: s//s")
    (s // s).plot(nfig="same")

    # Test substraction of Spectra
    s_tr = Transmittance_noslit(s)
    assert (s_tr - 1.0 * s_tr).get_integral("transmittance_noslit") == 0

    # TODO: add test
    # @EP: the test fails at the moment because multiply only works with radiance,
    # and MergeSlabs only works with non convoluted quantities
    # Do we want that? Up for discussion...

    # There should be an error if algebraic operations are used when
    # multiple quantities are defined:
    with pytest.raises(KeyError):
        2 * s

    s.apply_slit(0.1, "nm")
    s_rad = Radiance(s)

    # Test multiplication with float
    s.plot(lw=2, nfig="Multiplication (by scalar): 2*s", wunit="nm")
    #    (s*s).plot(nfig='same')
    (2 * s_rad).plot(nfig="same", wunit="nm")

    # Test algebraic addition (vs multiplication)
    assert (s_rad + s_rad).compare_with(2 * s_rad, spectra_only="radiance", plot=False)

    # Test algebraic addition with different waveunits
    s_rad_nm = s_rad.resample(s_rad.get_wavelength(), "nm", inplace=False)
    s_sum = 2 * s_rad_nm - s_rad_nm
    s_sum.compare_with(s_rad, spectra_only="radiance", plot=True)
    assert (s_rad_nm + s_rad_nm).compare_with(
        2 * s_rad, spectra_only="radiance", plot=True, rtol=1e-3
    )


@pytest.mark.fast
def test_TestBaseline(plot=False, *args, **kwargs):

    s = load_spec(getTestFile("CO_Tgas1500K_mole_fraction0.01.spec"), binary=True)
    s.update("radiance_noslit", verbose=False)
    s.apply_slit(0.1)
    s = Radiance_noslit(s)
    assert s.units["radiance"] == "mW/cm2/sr/nm"

    s2 = sub_baseline(s, 2e-4, -2e-4)
    if plot:
        plot_diff(s, s2)
    assert s2.get_radiance_noslit()[-1] == s.get_radiance_noslit()[-1] + 2e-4
    assert s2.get_radiance_noslit()[0] == s.get_radiance_noslit()[0] - 2e-4


@pytest.mark.fast
def test_dimensioned_operations(*args, **kwargs):

    import astropy.units as u
    import numpy as np

    from radis import Radiance, load_spec
    from radis.spectrum import sub_baseline
    from radis.test.utils import getTestFile

    # Generate the equivalent of an experimental spectrum
    s = load_spec(getTestFile(r"CO_Tgas1500K_mole_fraction0.01.spec"), binary=True)
    s.update()  # add radiance, etc.
    s.apply_slit(0.5)  # nm
    s = Radiance(s)

    # Test
    assert s.units["radiance"] == "mW/cm2/sr/nm"
    Imax = s.get("radiance", trim_nan=True)[1].max()

    # add a baseline
    s += 0.1 * u.Unit("W/cm2/sr/nm")

    assert np.isclose(s.get("radiance", trim_nan=True)[1].max(), Imax + 100)

    # remove a baseline (we could also have used s=-0.1, but we're testing another function here)
    s = sub_baseline(s, 0.1 * u.Unit("W/cm2/sr/nm"), 0.1 * u.Unit("W/cm2/sr/nm"))

    assert np.isclose(s.get("radiance", trim_nan=True)[1].max(), Imax)

    # Test division
    # Example : a manual normalization
    s /= s.max() * u.Unit("mW/cm2/sr/nm")

    assert s.units["radiance"] == ""  # normalized
    assert s.max() == 1.0

    # Test Multiplication
    # example : a manual Calibration
    s.units["radiance"] = "count"
    s *= 100 * u.Unit("mW/cm2/sr/nm/count")

    assert u.Unit(s.units["radiance"]) == u.Unit(
        "mW/cm2/sr/nm"
    )  # check units are valid
    assert s.units["radiance"] == "mW / (cm2 nm sr)"  # check units have been simplified


def _run_testcases(verbose=True, plot=False, *args, **kwargs):
    """Test procedures"""

    test_crop(verbose=verbose, *args, **kwargs)
    test_cut_recombine(verbose=verbose, *args, **kwargs)
    test_invariants(verbose=verbose, *args, **kwargs)
    test_operations_inplace(verbose=verbose, *args, **kwargs)
    test_serial_operator(verbose=verbose, plot=plot, *args, **kwargs)
    test_multiplyAndAddition(verbose=verbose, plot=plot, *args, **kwargs)
    test_offset(verbose=verbose, plot=plot, *args, **kwargs)
    test_other_algebraic_operations(verbose=verbose, plot=plot, *args, **kwargs)
    test_TestBaseline(verbose=verbose, plot=plot, *args, **kwargs)
    test_dimensioned_operations(*args, **kwargs)

    return True


if __name__ == "__main__":

    _run_testcases(plot=True)



# -*- coding: utf-8 -*-
"""Created on Wed Nov  5 12:59:37 2014.

@author: Erwan
Small functions used in other procedures
-------------------------------------------------------------------------------
"""


import os
import sys
from io import StringIO
from itertools import filterfalse, tee
from os.path import abspath, join, normcase, normpath

import numpy as np
import pandas as pd

verbose = True

# %%
# ==============================================================================
#  OS Functions
# ==============================================================================


def make_folders(path, folders):
    """Make folders if not there
    Parameters
    ----------
    path: str
        where to create folders
    folders: list or str
        folders to create
    """
    if type(folders) is str:
        folders = [folders]
    for folder in folders:
        try:
            os.mkdir(join(path, folder))
        except OSError:
            pass


# %%
# ==============================================================================
# Basic Functions
# ==============================================================================


def all_in(keys, L):
    """Returns whether all items in keys are in list L."""
    return all([k in L for k in keys])


def in_all(key, list_of_list):
    """Returns true if key is in all lists."""
    return all([key in L for L in list_of_list])


def any_in(keys, L):
    """Returns whether any of the items in keys are in list L."""
    return any([k in L for k in keys])


def key_max_val(d):
    """Return the dictionary key with max value."""
    v = list(d.values())
    k = list(d.keys())
    return k[v.index(max(v))]


def exec_file(afile, globalz=None, localz=None):
    with open(afile, "r") as fh:
        exec(fh.read(), globalz, localz)


def remove_duplicates(l):
    """Remove duplicates from a list, without changing the order.

    Note that if the order doesn't matter you could just do set(l)
    """

    l1 = []
    for e in l:
        if not e in l1:
            l1.append(e)
    return l1


def partition(pred, iterable):
    """Use a predicate to partition entries into false entries and true entries
    Returns
    -------
    Returns two lists: positive, and negative
    Example
    -------
     >>> partition(is_odd, range(10))
     --> [0 2 4 6 8], [1 3 5 7 9]
    """
    t1, t2 = tee(iterable)
    return list(filter(pred, t2)), list(filterfalse(pred, t1))


# Just for performance comparison. Below: naive implementation of partition
# %timeit partition(lambda x:x>0.5, np.random.rand(10))
# ... partition         2.67 µs per loop
# ... partition_naive   28.5 µs per loop
# def partition_naive(pred, iterable):
#    trues = []
#    falses = []
#    for item in iterable:
#        if pred(item):
#            trues.append(item)
#        else:
#            falses.append(item)
#    return trues, falses

# %% Compare / merge tools


def compare_dict(
    d1,
    d2,
    verbose="if_different",
    compare_as_paths=[],
    return_string=False,
    df1_str="Left",
    df2_str="Right",
    ignore_keys=[],
):
    """Returns ratio of equal keys [0-1]
    If verbose, also print all keys and values on 2 columns
    Parameters
    ----------
    d1, d2: dict
        two dictionaries to compare

    Other Parameters
    ----------------
    compare_as_paths: list of keys
        compare the values corresponding to given keys as path (irrespective of
        forward / backward slashes, or case )
    verbose: boolean, or ``'if_different'``
        ``'if_different'`` means results will be shown only if there is a difference.
    return_string: boolean
        if ``True``, returns message instead of just printing it (useful in error messages)
        Default ``False``
    ignore_keys: list
        do not compare these keys

    Returns
    -------
    out: float [0-1]
        ratio of matching keys
    if ``return_string``:
    out, string: float [0-1], str
        ratio of matching keys and comparison message
    """

    old_stdout = sys.stdout
    sys.stdout = newstdout = StringIO()  # capture all print

    try:

        print("{0:15}{1}\t{2}".format("Key", df1_str, df2_str))
        print("-" * 40)
        all_keys = set(list(d1.keys()) + list(d2.keys()))
        all_keys = [k for k in all_keys if k not in ignore_keys]
        s = 0  # counter of all matching keys
        for k in all_keys:
            if k in d1 and k in d2:  # key in both dicts. Let's compare values
                # Deal with path case
                if k in compare_as_paths:
                    if not compare_paths(d1[k], d2[k]):
                        print("{0:15}{1}\t{2}".format(k, d1[k], d2[k]))
                    else:
                        s += 1
                # Other cases
                else:
                    if d1[k] != d2[k]:
                        print("{0:15}{1}\t{2}".format(k, d1[k], d2[k]))
                    else:
                        s += 1
            elif k in d1:
                print("{0:15}{1}\tN/A".format(k, d1[k]))
            else:
                print("{0:15}N/A\t{1}".format(k, d2[k]))
        print("-" * 40)

        if len(all_keys) == 0:
            out = 1
        else:
            out = s / len(all_keys)

        # Get string output
        string = newstdout.getvalue()
        sys.stdout = old_stdout  # reset normal print

        # Output
        if verbose == True or (verbose == "if_different" and out != 1):
            print(string)

        if return_string:
            out = out, string

        return out
    except:
        raise
    finally:
        sys.stdout = old_stdout


def compare_lists(l1, l2, verbose="if_different", return_string=False):
    """Compare 2 lists of elements that may not be of the same length, irrespective
    of order. Returns the ratio of elements [0-1] present in both lists. If verbose,
    prints the differences
    Parameters
    ----------
    l1, l2: list-like
    verbose: boolean, or 'if_different'
        'if_different' means results will be shown only if there is a difference.
        function is called twice
    Other Parameters
    ----------------
    verbose: boolean, or ``'if_different'``
        ``'if_different'`` means results will be shown only if there is a difference.
    return_string: boolean
        if ``True``, returns message instead of just printing it (useful in error messages)
        Default ``False``
    Returns
    -------
    out: float [0-1]
        ratio of matching keys
    """

    old_stdout = sys.stdout
    sys.stdout = newstdout = StringIO()  # capture all print

    try:

        print("{0:20}\t{1}".format("Left", "Right"))
        print("-" * 40)
        all_keys = set(list(l1) + list(l2))
        s = 0  # counter of all matching keys
        for k in all_keys:
            if k in l1 and k in l2:  # key in both lists
                s += 1
            elif k in l1:
                print("{0:20}\tN/A".format("{0} ({1})".format(k, type(k))))
            else:
                print("{0:20}\t{1} ({2})".format("N/A", k, type(k)))
        print("-" * 40)

        if len(all_keys) == 0:
            out = 1
        else:
            out = s / len(all_keys)

        # Get string output
        string = newstdout.getvalue()
        sys.stdout = old_stdout  # reset normal print

        # Output
        if verbose == True or (verbose == "if_different" and out != 1):
            print(string)

        if return_string:
            out = out, string

        return out
    except:
        raise
    finally:
        sys.stdout = old_stdout


def stdpath(p):
    """Convert path p in standard path (irrespective of slash / backslash, or
    case)"""

    return normpath(normcase(abspath(p)))


def compare_paths(p1, p2):
    """Compare 2 paths p1 and p2."""
    return stdpath(p1) == stdpath(p2)


def merge_lists(lists):
    """Merge a list of lists and return a list with unique elements."""
    return list(set(sum([l for l in lists], [])))


# ==============================================================================
# %% Pandas specific
# ==============================================================================


def merge_rename_columns(df, columns1, columns2, merged_names):
    """Merge all columns under easier names. Only keep the useful ones
    Returns a new dataframe
    Parameters
    ----------
    df: pandas Dataframe
    columns1: list
        list of columns names
    columns2: list
        list of columns names, whose index match columns 1
    merged_names: list
        new names
    Example
    -------
    df = merge_rename_columns(df1, ['lvl_u', 'ju', 'Eu', 'nu', 'gu', 'grotu'],
                                   ['lvl_l', 'jl', 'El', 'nl', 'gl', 'grotl'],
                                   ['lvl',   'j',  'E',  'n',  'g',  'grot']
                                   )
    """

    assert all_in(columns1, list(df.keys()))
    assert all_in(columns2, list(df.keys()))

    df1 = df.loc[:, columns1]
    df2 = df.loc[:, columns2]
    df1.rename(
        columns={columns1[i]: merged_names[i] for i in range(len(merged_names))},
        inplace=True,
    )
    df2.rename(
        columns={columns2[i]: merged_names[i] for i in range(len(merged_names))},
        inplace=True,
    )
    df = pd.concat((df1, df2), ignore_index=True)

    return df.drop_duplicates()


def print_series(a):
    """Print a pandas series `a` , explicitely showing all rows."""

    for i, k in enumerate(a.keys()):
        print(k, "\t", a.values[0][i])


def transfer_metadata(df1, df2, metadata):
    """Transfer metadata between a DataFrame df1 and df2.

    For some reason metadata are sometimes not copied when a DataFrame is
    sliced or copied, even if they explicitely appear in the ``df.attrs``
    attribute. See https://github.com/pandas-dev/pandas/issues/28283

    Here we copy them back. Attributes can be :
        - keys of the ``df1.attrs`` dictionary
        - simple attributes of ``df1``, i.e., ``df1.X``

    Parameters
    ----------
    df1: pandas DataFrame
        copy from df1
    df2: pandas DataFrame
        copy to df2
    """

    for k in metadata:
        if not hasattr(df2, k):
            assert k not in df1.columns  # point is not to copy columns!
            if k in df1.attrs:  # Keys of attribute dictionary
                df2.attrs[k] = df1.attrs[k]
            else:  # Direct attributes of df1
                setattr(df2, k, getattr(df1, k))

    # @dev: refactor : we're updating the database to properly store values
    # either in columns either in the .attrs dict, but so they can always
    # be accessed with df.X


def expand_metadata(df, metadata):
    """Turn metadata from a float to a column.

    For some reason metadata are sometimes not copied when a DataFrame is
    sliced or copied, even if they explicitely figure in the df.attrs
    attribute. Here we add them as column before such operations.

    Parameters
    ----------

    df: pandas DataFrame
        ...

    Returns
    -------

    None:
        df modified in place
    """

    for k in metadata:
        if __debug__ and k not in df.attrs:
            from radis.misc.debug import printdbg

            printdbg("WARNING. {0} not in metadata: {1}".format(k, df.attrs))
        df[k] = getattr(df, k)


# %%
# ==============================================================================
# Types
# ==============================================================================


def list_if_float(a):
    if type(a) is list:
        return a
    else:
        return [a]


def flatten(*args):
    """Flatten list of lists of floats."""
    out = []
    for a in args:
        if is_list(a):
            out += list(a)
        else:
            out += [a]
    return out


def is_list(a):
    """Returns True if a has list-like type: list, np.array, tuple, set,
    etc.)"""
    return type(a) in [list, np.ndarray, tuple, set]


def is_float(a):
    """Returns True if a has float-like type: float, np.float64, np.int64,
    etc.)"""
    return type(a) in [float, np.float64, np.int32, np.float32, int, np.int64]


def is_number(s):
    """Return True if ``s`` is a number.

    Works for strings, floats, int, and is compatible with Python 2/3
    """

    try:
        float(s)
        return True
    except ValueError:
        return False


def to_str(a):
    if isinstance(a, bytes):
        return a.decode("utf-8")
    else:
        return a


def round_off(n):
    # Getting rounded off value (atleast order 3)
    for i in range(0, 10):
        val = round(n, 3 + i)
        if val == 0:
            continue
        return val
    return 0



# -*- coding: utf-8 -*-
"""Physical constants.

Note : more units available in :
- scipy.constants
- sympy.physics.units


-------------------------------------------------------------------------------
"""

# %% SI units

eV = 1.60217657e-19  # J
"""float: electron volt (J)"""

h = 6.62606957e-34
"""float: Planck constant (m2.kg.s-1 = J.s)"""

k_b = 1.3806488e-23
"""float: Boltzmann constant (m2.kg.s-2.K-1)"""

c = 2.99792458e8
"""float: light velocity (m/s)"""

a_0 = 5.29177e-11
"""float: First Bohr radius (m)"""

m_e = 9.10938291e-31
"""float: Electron mass (kg)"""

Ry = 13.60569253
"""float: Rydberg unit of energy (eV)"""

Av = 6.02214129e23
"""float: Avogadro constant (molecules / mol)"""
Na = Av  # just an alias

eps_0 = 8.854187817e-12
"""float: Vacuum permittivity (Farads / m)"""

hc_k = h * c / k_b * 100  #
"""float:  cm to K conversion (~ 1.44 K/cm), used in Boltman factors"""


# %% HITRAN (CGS) units

k_b_CGS = 1.380648813e-16
"""float: Boltzman constant (erg / K)"""

c_CGS = 2.99792458e10
"""float: Light velocity (cm / s)"""

h_CGS = 6.626196e-27
"""float: Planck constant (erg * s)"""



# -*- coding: utf-8 -*-
"""
Summary
-------

Functions to parse the ``~/radis.json`` :ref:`Configuration file <label_lbl_config_file>`

Notes
-----

Create a ~/radis.json file in your HOME that contains all machine-specific information
(e.g: path to databanks). See :data:`~radis.misc.config.DBFORMAT` for expected
format

Routine Listing
---------------

- :func:`~radis.misc.config.get_user_config_configformat`
- :func:`~radis.misc.config.getDatabankEntries_configformat`
- :func:`~radis.misc.config.getDatabankList_configformat`
- :func:`~radis.misc.config.addDatabankEntries_configformat`
- :func:`~radis.misc.config.diffDatabankEntries`
- :func:`~radis.misc.config.printDatabankEntries_configformat`
- :func:`~radis.misc.config.printDatabankList_configformat`
- :func:`~radis.misc.config.get_user_config`
- :func:`~radis.misc.config.convertRadisToJSON`
- :func:`~radis.misc.config.getDatabankEntries`
- :func:`~radis.misc.config.getDatabankList`
- :func:`~radis.misc.config.addDatabankEntries`
- :func:`~radis.misc.config.printDatabankEntries`
- :func:`~radis.misc.config.printDatabankList`

-------------------------------------------------------------------------------


"""


import configparser
import json
import os
import shutil
import warnings
from json import JSONDecodeError
from os.path import dirname, exists, expanduser, join

from radis.misc.basics import compare_dict, compare_lists, stdpath
from radis.misc.utils import DatabankNotFound, getProjectRoot
from radis.misc.warning import DatabaseAlreadyExists

# %% Functions to parse radis/config.json

CONFIG_PATH_DEFAULT = join(getProjectRoot(), "default_radis.json")
CONFIG_PATH_JSON = join(expanduser("~"), "radis.json")
CONFIG_PATH_OLD = join(expanduser("~"), ".radis")

assert exists(CONFIG_PATH_DEFAULT)


def get_config(configpath=CONFIG_PATH_JSON):
    """Read the default RADIS config.json file `configpath` (default
    :py:attr:`~radis.misc.config.CONFIG_PATH_JSON` and override it with the
    entries of the user config file ``~/.radis``
    (:py:attr:`~radis.misc.config.CONFIG_PATH_DEFAULT`"""

    jsonfile = CONFIG_PATH_DEFAULT
    with open(jsonfile) as f:
        try:
            config = json.load(f)
        except JSONDecodeError as err:
            raise JSONDecodeError(
                "Error reading '{0}' (line {2} col {3}): \n{1}".format(
                    jsonfile, err.msg, err.lineno, err.colno
                ),
                err.doc,
                err.pos,
            ) from err

    user_config = get_user_config(configpath)

    config.update(user_config)

    return config


# %% Functions to parse ~/.radis file

DBFORMAT = r"""
--------------------------

[MY-HITEMP-CO2]                  #  your databank name: use this in calc_spectrum()
                                 #  or SpectrumFactory.load_databank()
info = HITEMP 2010 databank      #  whatever you want
path =                           #  no "", multipath allowed
       D:\Databases\HITEMP-CO2\hitemp_07
       D:\Databases\HITEMP-CO2\hitemp_08
       D:\Databases\HITEMP-CO2\hitemp_09
format = hitran                  #  'hitran' (HITRAN/HITEMP), 'cdsd-hitemp', 'cdsd-4000'
                                 # databank text file format. More info in
                                 # SpectrumFactory.load_databank function.
parfuncfmt:                      #  'cdsd', 'hapi', etc.
                                 # format to read tabulated partition function
                                 # file. If `hapi`, then HAPI (HITRAN Python
                                 # interface) is used to retrieve them (valid if
                                 # your databank is HITRAN data). HAPI is embedded
                                 # into RADIS. Check the version.
# Optional
# ----------
parfunc:                         #  path to tabulated partition function to use.
                                 # If `parfuncfmt` is `hapi` then `parfunc`
                                 # should be the link to the hapi.py file. If
                                 # not given, then the hapi.py embedded in RADIS
                                 # is used (check version)
levels_iso1                      #  path to energy levels (needed for non-eq
                                 # calculations). Default None
levels_iso2                      # etc
levels_iso4                      # etc
levelsfmt:                       #  'cdsd', etc.
                                 # how to read the previous file. Default None.
levelsZPE:                       #  zero-point-energy (cm-1): offset for all level
                                 # energies. Default 0 (if not given)

--------------------------"""
"""str: Typical expected format of a ~/.radis entry::

    --------------------------

    [MY-HITEMP-CO2]                  #  your databank name: use this in calc_spectrum()
                                     #  or SpectrumFactory.load_databank()
    info = HITEMP 2010 databank      #  whatever you want
    path =                           #  no "", multipath allowed
           D:\Databases\HITEMP-CO2\hitemp_07
           D:\Databases\HITEMP-CO2\hitemp_08
           D:\Databases\HITEMP-CO2\hitemp_09
    format = hitran                  #  'hitran' (HITRAN/HITEMP), 'cdsd-hitemp', 'cdsd-4000'
                                     # databank text file format. List of all
                                     # formats in :py:data:`~radis.lbl.loader.KNOWN_DBFORMAT`
                                     # More info in
                                     # :py:meth:`~radis.lbl.loader.DatabankLoader.load_databank` function.
    parfuncfmt:                      #  'cdsd', 'hapi', etc.
                                     # format to read tabulated partition function
                                     # file. If `hapi`, then HAPI (HITRAN Python
                                     # interface) is used to retrieve them (valid if
                                     # your databank is HITRAN data). HAPI is embedded
                                     # into RADIS. Check the version.
                                     # List of all formats in :py:data:`~radis.lbl.loader.KNOWN_LVLFORMAT`
    # Optional
    # ----------
    parfunc:                         #  path to tabulated partition function to use.
                                     # If `parfuncfmt` is `hapi` then `parfunc`
                                     # should be the link to the hapi.py file. If
                                     # not given, then the hapi.py embedded in RADIS
                                     # is used (check version)
    levels_iso1                      #  path to energy levels (needed for non-eq
                                     # calculations). Default None
    levels_iso2                      # etc
    levels_iso4                      # etc
    levelsfmt:                       #  'cdsd', etc.
                                     # how to read the previous file. Default None.
    levelsZPE:                       #  zero-point-energy (cm-1): offset for all level
                                     # energies. Default 0 (if not given)

    --------------------------

For more information refer to the documentation: :ref:`Configuration file <label_lbl_config_file>` :

- Setup test databases with :py:func:`~radis.test.utils.setup_test_line_databases`
- `format` :  format: :py:data:`~radis.lbl.loader.KNOWN_DBFORMAT`


See Also
--------

:ref:`Configuration file <label_lbl_config_file>`,
:py:func:`~radis.misc.config.get_user_config`,
:py:meth:`~radis.lbl.loader.DatabankLoader.load_databank`
"""

DBFORMATJSON = r"""
--------------------------

{
"database": {                       # database key, stores all databases
    "MY-HITEMP-CO2": {              # your databank name: use this in calc_spectrum()
                                    # or SpectrumFactory.load_databank()
    "info": "HITEMP 2010 databank", # whatever you want
    "path":                         # no "", multipath allowed
       ["D:\Databases\HITEMP-CO2\hitemp_07",
        "D:\Databases\HITEMP-CO2\hitemp_08",
        "D:\Databases\HITEMP-CO2\hitemp_09"],
    "format": "hitran",             # 'hitran' (HITRAN/HITEMP), 'cdsd-hitemp', 'cdsd-4000'
                                    # databank text file format. More info in
                                    # SpectrumFactory.load_databank function.
    "parfuncfmt":                   # 'cdsd', 'hapi', etc.
                                    # format to read tabulated partition function
                                    # file. If `hapi`, then HAPI (HITRAN Python
                                    # interface) is used to retrieve them (valid if
                                    # your databank is HITRAN data). HAPI is embedded
                                    # into RADIS. Check the version.
    # Optional
    # ----------
    "parfunc":                      # path to tabulated partition function to use.
                                    # If `parfuncfmt` is `hapi` then `parfunc`
                                    # should be the link to the hapi.py file. If
                                    # not given, then the hapi.py embedded in RADIS
                                    # is used (check version)
    "levels_iso1":                  # path to energy levels (needed for non-eq
                                    # calculations). Default None
    "levels_iso2":                  # etc
    "levels_iso4":                  # etc
    "levelsfmt":                    # 'cdsd', etc.
                                    # how to read the previous file. Default None.
    "levelsZPE":                    # zero-point-energy (cm-1): offset for all level
    }                               # energies. Default 0 (if not given)
  }
}

--------------------------"""
"""str: Typical expected format of a ~/radis.json entry in "database" key::

    --------------------------

    "MY-HITEMP-CO2": {              # your databank name: use this in calc_spectrum()
                                    # or SpectrumFactory.load_databank()
    "info": "HITEMP 2010 databank", # whatever you want
    "path":[                        # no "", multipath allowed
        "D:\Databases\HITEMP-CO2\hitemp_07"
        "D:\Databases\HITEMP-CO2\hitemp_08"
        "D:\Databases\HITEMP-CO2\hitemp_09"
    ],
    "format": "hitran",             # 'hitran' (HITRAN/HITEMP), 'cdsd-hitemp', 'cdsd-4000'
                                    # databank text file format. List of all
                                    # formats in :py:data:`~radis.lbl.loader.KNOWN_DBFORMAT`
                                    # More info in
                                    # :py:meth:`~radis.lbl.loader.DatabankLoader.load_databank` function.
    "parfuncfmt":                   # 'cdsd', 'hapi', etc.
                                    # format to read tabulated partition function
                                    # file. If `hapi`, then HAPI (HITRAN Python
                                    # interface) is used to retrieve them (valid if
                                    # your databank is HITRAN data). HAPI is embedded
                                    # into RADIS. Check the version.
                                    # List of all formats in :py:data:`~radis.lbl.loader.KNOWN_LVLFORMAT`
    # Optional
    # ----------
    "parfunc":                      # path to tabulated partition function to use.
                                    # If `parfuncfmt` is `hapi` then `parfunc`
                                    # should be the link to the hapi.py file. If
                                    # not given, then the hapi.py embedded in RADIS
                                    # is used (check version)
    "levels_iso1":                  # path to energy levels (needed for non-eq
                                    # calculations). Default None
    "levels_iso2":                  # etc
    "levels_iso4":                  # etc
    "levelsfmt":                    # 'cdsd', etc.
                                    # how to read the previous file. Default None.
    "levelsZPE":                    # zero-point-energy (cm-1): offset for all level
    }                               # energies. Default 0 (if not given)

    --------------------------

For more information refer to the documentation: :ref:`Configuration file <label_lbl_config_file>` :

- Setup test databases with :py:func:`~radis.test.utils.setup_test_line_databases`
- `format` :  format: :py:data:`~radis.lbl.loader.KNOWN_DBFORMAT`


See Also
--------

:ref:`Configuration file <label_lbl_config_file>`,
:py:func:`~radis.misc.config.get_user_config`,
:py:meth:`~radis.lbl.loader.DatabankLoader.load_databank`
"""


def get_user_config_configformat():
    """Read config file and returns it.

    Config file name is harcoded: :ref:`~/.radis <label_lbl_config_file>`
    """

    config = configparser.ConfigParser()
    configpath = CONFIG_PATH_OLD

    # Test ~/.radis exists
    if not exists(configpath):

        raise FileNotFoundError(
            "Create a `.radis` file in {0} to store links to ".format(
                dirname(configpath)
            )
            + "your local databanks. Format must be:\n {0}".format(DBFORMAT)
        )
    config.read(configpath)

    return config


def convertRadisToJSON(config_path_json, config_path_old=CONFIG_PATH_OLD):
    """Converts the ~/.radis file into json formatted file ~/radis.json

    Example
    -------
    original ~/.radis file format::

        [HITRAN-CO2-TEST]
        info = HITRAN 2016 database, CO2, 1 main isotope (CO2-626), bandhead: 2380-2398 cm-1 (4165-4200 nm)
        path = PATH_TO\radis\radis\test\files\hitran_co2_626_bandhead_4165_4200nm.par
        format = hitran
        parfuncfmt = hapi
        levelsfmt = radis

    -----------
    Converted ~/radis.json file format::

        {
        "database": {
            "HITRAN-CO2-TEST": {
                "info": "HITRAN 2016 database, CO2, 1 main isotope (CO2-626), bandhead: 2380-2398 cm-1 (4165-4200 nm)",
                "path": "PATH_TO\\radis\\radis\\test\\files\\hitran_co2_626_bandhead_4165_4200nm.par",
                "format": "hitran",
                "parfuncfmt": "hapi",
                "levelsfmt": "radis"
            }
          }
        }
    """

    # Loads configuration file ~/.radis
    config = get_user_config_configformat(config_path_old)

    # Variable to store data in JSON format
    config_json = {}

    # Converting configuration into JSON format and storing in config_json variable
    for i in config.sections():
        temp = {}
        for j in config[i]:
            # Checking if the `path` is multiple then creating list otherwise string
            if j == "path":
                if "\n" in config[i][j]:
                    store_list = config[i][j].split("\n")
                    # using remove() to remove empty list elements
                    while "" in store_list:
                        store_list.remove("")

                    if len(store_list) == 1:
                        temp[j] = store_list[0]
                    else:
                        temp[j] = store_list
                else:
                    temp[j] = config[i][j]
            else:
                # Adding to `temp` dictionaru
                temp[j] = config[i][j]

        config_json[i] = temp

    # Adding all config element to a `database` key
    config_final = {}
    config_final["database"] = config_json

    # Creating json file
    config_json_dir = config_path_json
    with open(config_json_dir, "w") as outfile:
        json.dump(config_final, outfile, indent=3)
    outfile.close()

    return


def init_radis_json(config_path_json, config_path_old=CONFIG_PATH_OLD):
    """Checks whether ``config_path_json`` (usually `~/radis.json`) exists.

    If not then we use
    :func:`~radis.misc.config.convertRadisToJSON`
    to convert `~/.radis` into `~/radis.json`
    """
    # If `~/.radis` and `~/radis.json` both found, raises DeprecationWarning
    if exists(config_path_json) and exists(config_path_old):
        raise (
            DeprecationWarning(
                f"{config_path_old} and {config_path_json} both found. "
                + "`~/.radis` config file was replaced by `~/radis.json` in version '0.9.29'."
                + f" Make sure all file content is in {config_path_json} then remove {config_path_old}.",
            )
        )
    elif exists(config_path_json):
        pass
    elif exists(config_path_old):
        convertRadisToJSON(config_path_json, config_path_old)
    else:
        # Create it from the default config path
        try:
            shutil.copy2(CONFIG_PATH_DEFAULT, config_path_json)
        except Exception as err:
            print(
                f"Couldn't create RADIS configuration in {config_path_json} ({str(err)})"
            )
        else:
            print(f"RADIS configuration file created in {config_path_json}")

    # Check it user file exists
    if not exists(config_path_json):
        raise FileNotFoundError(
            "Create a `radis.json file in {0} to store links to ".format(
                dirname(config_path_json)
            )
            + "your local databanks. Format must be:\n {0}".format(DBFORMATJSON)
            + "\n(it can be empty too)"
        )

    return


def get_user_config(configpath=CONFIG_PATH_JSON):
    """Read config file and returns it."""

    # First checking `~radis.json` exist or not, if not create it
    init_radis_json(configpath)

    # Checking file is empty
    if os.stat(configpath).st_size == 0:
        return {}

    # Load the JSON file
    with open(configpath) as f:
        try:
            config = json.load(f)
        except JSONDecodeError as err:
            raise JSONDecodeError(
                "Error reading '{0}' (line {2} col {3}): \n{1}".format(
                    configpath, err.msg, err.lineno, err.colno
                ),
                err.doc,
                err.pos,
            ) from err

    return config


def createConfigFile(config_path, verbose=True):

    # Safety check : file shouldnt exist
    if exists(config_path):
        raise FileExistsError(
            f"Config file already exist: {config_path}. You shouldnt be re-creating it. If sure, check it and delete it manually"
        )

    # generate ~/radis.json:
    open(config_path, "w").close()
    if verbose:
        print("Created ~/radis.json in {0}".format(dirname(config_path)))


def getDatabankEntries_configformat(dbname, get_extra_keys=[]):
    r"""Read :ref:`~/.radis <label_lbl_config_file>` config file and returns a dictionary of entries.

    Parameters
    ----------
    dbname: str
        database name in :ref:`~/.radis <label_lbl_config_file>` config file
    get_extra_keys: list
        read additional parameters on top of the usual Databank format keys :

    Notes
    -----
    Databank format::

        [MY-HITEMP-CO2]                  #  your databank name: use this in calc_spectrum()
                                         #  or SpectrumFactory.load_databank()
        info = HITEMP 2010 databank      # whatever you want
        path =                           # no "", multipath allowed
               D:\Databases\HITEMP-CO2\hitemp_07
               D:\Databases\HITEMP-CO2\hitemp_08
               D:\Databases\HITEMP-CO2\hitemp_09
        format = hitemp                  # 'hitran' (HITRAN / HITEMP), 'cdsd-hitemp', 'cdsd-4000'
                                         # Databank text file format. More info in
                                         # SpectrumFactory.load_databank function.

        # Optional:

        parfunc                          # path or 'USE_HAPI'
                                         # path to tabulated partition functions. If
                                         # `USE_HAPI`, then HAPI (HITRAN Python
                                         interface) [1]_ is used to retrieve them (valid
                                         if your databank is HITRAN data). HAPI
                                         is embedded into RADIS. Check the version.

        parfuncfmt:                      # 'cdsd'
                                         # format to read tabulated partition function
                                         # file. If `USE_HAPI` is given as `parfunc`
                                         # parameter then this line should not be used.

        levels_iso1                      # path to energy levels (needed for non-eq)
                                         # calculations.
        levels_iso2                      # etc
        levels_iso4                      # etc

        levelsfmt                        # 'cdsd'
                                         # how to read the previous file.

    References
    ----------
    .. [1] `HAPI: The HITRAN Application Programming Interface <http://hitran.org/hapi>`_


    """
    warnings.warn(DeprecationWarning("config format changed to JSON in radis 0.9.29"))
    config = get_user_config_configformat()

    # Make sure it looks like a databank (path and format are given)
    try:
        config.get(dbname, "path")
        config.get(dbname, "format")
    except configparser.NoSectionError:
        msg = (
            "{1}\nDBFORMAT\n{0}\n".format(DBFORMAT, dbname)
            + "No databank named {0} in `{1}`. ".format(dbname, CONFIG_PATH_OLD)
            + "Available databanks: {0}. ".format(getDatabankList_configformat())
            + "See databank format above. More information in "
            + "https://radis.readthedocs.io/en/latest/lbl/lbl.html#configuration-file"
        )
        raise DatabankNotFound(msg)

    entries = dict(config.items(dbname))

    # Parse paths correctly
    entries["path"] = entries["path"].strip("\n").split("\n")

    # Merge all isotope-dependant levels into one dict
    iso_list = [k for k in entries if k.startswith("levels_iso")]
    if len(iso_list) > 0:
        levels = {}
        for k in iso_list:
            iso = float(k[10:])
            levels[iso] = entries.pop(k)
        entries["levels"] = levels
    else:
        if "levels" in entries:
            raise SyntaxError(
                "in {0}: `levels` replaced with ".format(CONFIG_PATH_OLD)
                + "`levels_iso#` where # is the isotope number"
            )

    return entries


def getDatabankList_configformat():
    """Get all databanks available in :ref:`~/.radis <label_lbl_config_file>`."""
    warnings.warn(DeprecationWarning("config format changed to JSON in radis 0.9.29"))

    config = get_user_config_configformat()

    # Get databank path and format
    validdb = []
    for dbname in config.sections():
        try:
            config.get(dbname, "path")
            config.get(dbname, "format")
        except configparser.NoSectionError:
            # not a db
            continue
        except configparser.NoOptionError:
            # not a db
            continue
        # looks like a db. add to db
        validdb.append(dbname)
    return validdb


def addDatabankEntries_configformat(dbname, dict_entries, verbose=True):
    """Add database dbname with entries from dict_entries.

    If database already exists in :ref:`~/.radis <label_lbl_config_file>`, raises an error

    Examples
    --------

    ::

        addDatabankEntries("HITEMP2010-CO2",
            {
                "info": "HITEMP2020 CO2 lines with TIPS-2017 for partition functions (equilibrium) and RADIS for rovibrational energies (nonequilibrium) ",
                "path": ["PATH/TO/HITEMP/CO2/*.par"],
                "format": "hitran",
                "parfuncfmt": "hapi",
                "levelsfmt": "radis",
            })
    """
    warnings.warn(DeprecationWarning("config format changed to JSON in radis 0.9.29"))

    # Get ~/.radis if exists, else create it
    try:
        dbnames = getDatabankList_configformat()
    except FileNotFoundError:
        # generate ~/.radis:
        dbnames = []
        open(CONFIG_PATH_OLD, "a").close()
        if verbose:
            print("Created ~/.radis in {0}".format(dirname(CONFIG_PATH_OLD)))

    # Check database doesnt exist
    if dbname in dbnames:
        raise DatabaseAlreadyExists(
            "Database already exists: {0}".format(dbname) + ". Cant add it"
        )

    # Add entries to parser
    config = configparser.ConfigParser()
    config[dbname] = {}

    if "info" in dict_entries:  # optional
        config[dbname]["info"] = dict_entries.pop("info")

    # ... Parse paths correctly
    if isinstance(dict_entries["path"], str):
        config[dbname]["path"] = dict_entries.pop("path")
    else:  # list
        config[dbname]["path"] = "\n       ".join(dict_entries.pop("path"))

    config[dbname]["format"] = dict_entries.pop("format")
    config[dbname]["parfuncfmt"] = dict_entries.pop("parfuncfmt")

    # Optional:
    # ... partition functions:
    if "parfunc" in dict_entries:
        config[dbname]["parfunc"] = dict_entries.pop("parfunc")

    # ... Split all isotopes in separate keys
    levels_dict = dict_entries.pop("levels", {})
    for iso, levels_iso in levels_dict.items():
        config[dbname]["levels_iso{0}".format(iso)] = levels_iso

    if "levelsfmt" in dict_entries:
        config[dbname]["levelsfmt"] = dict_entries.pop("levelsfmt")

    # Allow some informative parameters
    # ... "download_url", "download_date", "wavenumber_min", "wavenumber_max" generated by fetch_hitemp
    for entry in ["download_url", "download_date", "wavenumber_min", "wavenumber_max"]:
        if entry in dict_entries:
            config[dbname][entry] = dict_entries.pop(entry)

    # Check nothing is left
    if dict_entries != {}:
        raise ValueError("Unexpected keys: {0}".format(list(dict_entries.keys())))

    # Write to ~/.radis
    # ... Note: what if there is a PermissionError here? Try/except pass?
    with open(CONFIG_PATH_OLD, "a") as configfile:
        configfile.write("\n")
        config.write(configfile)
    if verbose:
        print("Added {0} database in {1}".format(dbname, CONFIG_PATH_OLD))

    return


def diffDatabankEntries(dict_entries1, dict_entries2, verbose=True):
    """Compare two Databank entries under dict format (i.e: output of
    getDatabankEntries)

    Returns None if no differences are found, or the first different key
    """

    k = None

    try:
        verbose_compare = "if_different" if verbose else False
        assert len(dict_entries1) == len(dict_entries2)
        assert (
            compare_lists(
                list(dict_entries1.keys()),
                list(dict_entries2.keys()),
                verbose=verbose_compare,
            )
            == 1
        )
        for k in dict_entries1.keys():
            v1 = dict_entries1[k]
            v2 = dict_entries2[k]
            if k in ["info", "format", "parfunc", "parfuncfmt", "levelsfmt"]:
                assert v1.lower() == v2.lower()
            elif k in ["path"]:
                assert (
                    compare_lists(
                        [stdpath(path1).lower() for path1 in v1],
                        [stdpath(path2).lower() for path2 in v2],
                        verbose=verbose_compare,
                    )
                    == 1
                )
            elif k in ["levels"]:
                assert (
                    compare_dict(
                        v1,
                        v2,
                        compare_as_paths=list(v1.keys()),
                        verbose=verbose_compare,
                    )
                    == 1
                )
            else:
                raise ValueError("Unexpected key:", k)

        return None

    except AssertionError:
        if verbose:
            print("Key doesnt match:", k)
        return k


def printDatabankEntries_configformat(dbname, crop=200):
    """Print databank info.

    Parameters
    ----------
    dbname: str
        database name in :ref:`~/.radis <label_lbl_config_file>`
    crop: int
        if > 0, cutoff entries larger than that
    """
    warnings.warn(DeprecationWarning("config format changed to JSON in radis 0.9.29"))
    entries = getDatabankEntries_configformat(dbname)
    print(dbname, "\n-------")
    for k, v in entries.items():
        # Add extra arguments
        args = []
        if k == "levelszpe":
            args.append("cm-1")
        v = "{0}".format(v)
        if len(v) > crop and crop > 0:
            v = v[:crop] + "..."
        # Print item
        print(k, ":", v, *args)


def printDatabankList_configformat():
    """Print all databanks available in ~/.radis"""
    warnings.warn(DeprecationWarning("config format changed to JSON in radis 0.9.29"))
    try:
        print(
            "Databanks in {0}: ".format(CONFIG_PATH_OLD),
            ",".join(getDatabankList_configformat()),
        )
        for dbname in getDatabankList_configformat():
            print("\n")
            printDatabankEntries_configformat(dbname)
    except FileNotFoundError:
        print("No config file {0}".format(CONFIG_PATH_OLD))
        # it's okay


def getDatabankEntries(dbname, get_extra_keys=[], configpath=CONFIG_PATH_JSON):
    r"""Read :ref:`~/radis.json <label_lbl_config_file>` config file and returns a dictionary of entries.

    Parameters
    ----------
    dbname: str
        database name in :ref:`~/radis.json <label_lbl_config_file>` config file
    get_extra_keys: list
        read additional parameters on top of the usual Databank format keys :

    Notes
    -----
    Databank format::

        "MY-HITEMP-CO2": {              # your databank name: use this in calc_spectrum()
                                        # or SpectrumFactory.load_databank()
        "info": "HITEMP 2010 databank"  # whatever you want
        "path": [                       # no "", multipath allowed
            "D:\Databases\HITEMP-CO2\hitemp_07",
            "D:\Databases\HITEMP-CO2\hitemp_08",
            "D:\Databases\HITEMP-CO2\hitemp_09"
        ],
        "format": "hitemp"              # 'hitran' (HITRAN / HITEMP), 'cdsd-hitemp', 'cdsd-4000'
                                        # Databank text file format. More info in
                                        # SpectrumFactory.load_databank function.

        # Optional:

        "parfunc":                      # path or 'USE_HAPI'
                                        # path to tabulated partition functions. If
                                        # `USE_HAPI`, then HAPI (HITRAN Python
                                        interface) [1]_ is used to retrieve them (valid
                                        if your databank is HITRAN data). HAPI
                                        is embedded into RADIS. Check the version.

        "parfuncfmt":                   # 'cdsd'
                                        # format to read tabulated partition function
                                        # file. If `USE_HAPI` is given as `parfunc`
                                        # parameter then this line should not be used.

        "levels_iso1":                  # path to energy levels (needed for non-eq)
                                        # calculations.
        "levels_iso2":                  # etc
        "levels_iso4":                  # etc

        "levelsfmt":                    # 'cdsd'
        }                               # how to read the previous file.



    References
    ----------
    .. [1] `HAPI: The HITRAN Application Programming Interface <http://hitran.org/hapi>`_


    """
    # Loading `~/radis.json` file
    _config = get_user_config(configpath)

    # Make sure it looks like a databank (path and format are given)
    try:
        # Getting the databases
        config = _config["database"]
        config[dbname]["path"]
        config[dbname]["format"]
    except KeyError:
        msg = (
            "{1}\nDBFORMAT\n{0}\n".format(DBFORMAT, dbname)
            + "No databank named {0} in `{1}`. ".format(dbname, configpath)
            + "Available databanks: {0}. ".format(
                getDatabankList(configpath=configpath)
            )
            + "See databank format above. More information in "
            + "https://radis.readthedocs.io/en/latest/lbl/lbl.html#configuration-file"
        )
        raise DatabankNotFound(msg)

    entries = config[dbname]

    # Path correction for all tests
    if type(entries["path"]) == str:
        entries["path"] = [entries["path"]]

    # Merge all isotope-dependant levels into one dict
    iso_list = [k for k in entries if k.startswith("levels_iso")]
    if len(iso_list) > 0:
        levels = {}
        for k in iso_list:
            iso = float(k[10:])
            levels[iso] = entries.pop(k)
        entries["levels"] = levels
    else:
        if "levels" in entries:
            raise SyntaxError(
                "in {0}: `levels` replaced with ".format(configpath)
                + "`levels_iso#` where # is the isotope number"
            )

    return entries


def getDatabankList(configpath=CONFIG_PATH_JSON):
    """Get all databanks available in :ref:`~/radis.json"""

    _config = get_user_config(configpath)

    # If `_config` is empty dictionary then return empty list
    if len(_config) == 0:
        return []

    # Getting the databases
    config = _config["database"]

    # Get databank path and format
    validdb = []
    for dbname in config:
        try:
            config[dbname]["path"]
            config[dbname]["format"]
        except configparser.NoSectionError:
            # not a db
            continue
        except configparser.NoOptionError:
            # not a db
            continue
        # looks like a db. add to db
        validdb.append(dbname)
    return validdb


def addDatabankEntries(dbname, dict_entries, verbose=True, configpath=CONFIG_PATH_JSON):
    """Add database dbname with entries from dict_entries.

    If database already exists in :ref:`~/radis.json <label_lbl_config_file>`, raises an error

    Examples
    --------

    ::

        addDatabankEntries("HITEMP2010-CO2",
            {
                "info": "HITEMP2020 CO2 lines with TIPS-2017 for partition functions (equilibrium) and RADIS for rovibrational energies (nonequilibrium) ",
                "path": ["PATH/TO/HITEMP/CO2/*.par"],
                "format": "hitran",
                "parfuncfmt": "hapi",
                "levelsfmt": "radis",
            })
    """

    # Get ~/radis.json if exists, else create it
    dbnames = getDatabankList(configpath)

    # Check database doesnt exist
    if dbname in dbnames:
        raise DatabaseAlreadyExists(
            f"Database {dbname} already exists in {configpath}. Cant add it"
        )

    # Loading `~/radis.json`
    with open(configpath, "r") as json_file:
        _config = json.load(json_file)

    try:
        # Accessing `database` key in file
        config = _config["database"]
    except KeyError:
        # Creating `database` key
        _config["database"] = {}
        config = _config["database"]

    # Adding entries in `config[dbname]`
    config[dbname] = {}

    if "info" in dict_entries:  # optional
        config[dbname]["info"] = dict_entries.pop("info")

    # ... Parse paths correctly
    if isinstance(dict_entries["path"], str):
        config[dbname]["path"] = dict_entries.pop("path")
    else:  # list
        # If list has a single path
        if len(dict_entries["path"]) == 1:
            config[dbname]["path"] = dict_entries.pop("path")[0]
        else:
            config[dbname]["path"] = dict_entries.pop("path")

    config[dbname]["format"] = dict_entries.pop("format")
    config[dbname]["parfuncfmt"] = dict_entries.pop("parfuncfmt")

    # Optional:
    # ... partition functions:
    if "parfunc" in dict_entries:
        config[dbname]["parfunc"] = dict_entries.pop("parfunc")

    # ... Split all isotopes in separate keys
    levels_dict = dict_entries.pop("levels", {})
    for iso, levels_iso in levels_dict.items():
        config[dbname]["levels_iso{0}".format(iso)] = levels_iso

    if "levelsfmt" in dict_entries:
        config[dbname]["levelsfmt"] = dict_entries.pop("levelsfmt")

    # Allow some informative parameters
    # ... "download_url", "download_date", "wavenumber_min", "wavenumber_max" generated by fetch_hitemp
    for entry in ["download_url", "download_date", "wavenumber_min", "wavenumber_max"]:
        if entry in dict_entries:
            config[dbname][entry] = dict_entries.pop(entry)

    # Check nothing is left
    if dict_entries != {}:
        raise ValueError("Unexpected keys: {0}".format(list(dict_entries.keys())))

    # Write to `~/radis.json`
    with open(configpath, "w") as configfile:
        json.dump(_config, configfile, indent=3)
        configfile.close()

    if verbose:
        print("Added {0} database in {1}".format(dbname, configpath))

    return


def printDatabankEntries(dbname, crop=200, configpath=CONFIG_PATH_JSON):
    """Print databank info.

    Parameters
    ----------
    dbname: str
        database name in :ref:`~/radis.json <label_lbl_config_file>`
    crop: int
        if > 0, cutoff entries larger than that
    """
    entries = getDatabankEntries(dbname, configpath=configpath)
    print("'{0}':".format(dbname))
    print(entries, "\n")


def printDatabankList(configpath=CONFIG_PATH_JSON):
    """Print all databanks available in ~/radis.json"""
    try:
        print(
            "Databanks in {0}: ".format(configpath),
            ",".join(getDatabankList(configpath=configpath)),
        )
        for dbname in getDatabankList(configpath=configpath):
            print("\n")
            printDatabankEntries(dbname, configpath=configpath)
    except FileNotFoundError:
        print("No config file {0}".format(configpath))
        # it's okay


# %% Test

if __name__ == "__main__":

    # from radis.test.misc.test_config import _run_testcases

    # _run_testcases(verbose=True)

    a = get_config()

    print(a.keys())
